(** * Performances Tests of the Frequent Itemset Mining Algorithm *)

Require Import Coq.Lists.List ZArith. Import ListNotations.
Require Import SyDPaCC.FIM.Support.List SyDPaCC.FIM.SeqFIM
        SyDPaCC.FIM.DataZoo.

Module Numbers.
  Definition A := Z.
  Definition eq_dec := Z.eq_dec.

  Definition leb := Z.leb.
  Definition le := Z.le.


  
  Lemma le_refl : forall x, le x x.
    unfold le; intros; auto with zarith.
  Qed.
  
  Lemma le_trans : forall x y z, le x y -> le y z -> le x z.
    unfold le; intros; auto with zarith.
  Qed.
  
  Lemma le_antisym : forall x y, le x y -> le y x -> x = y.
    unfold le; intros; auto with zarith.
  Qed.
  
  Lemma le_dec : forall x y, {le x y} + {~ (le x y)}.
    unfold le; apply Z_le_dec.
  Qed.
    
  Lemma leb_le : forall n m : A, (leb n m) = true <-> (le n m).
    unfold leb, A. split; apply Zle_is_le_bool.
  Qed.
    
  Lemma leb_nle : forall n m : A, (leb n m) = false <-> not (le n m).
    unfold leb, A. apply Z.leb_nle.
  Qed.
    
End Numbers.

Module F  := Fim Numbers.
Module LO := List_Order Numbers Numbers. 
Module FO := FimOrder Numbers Numbers LO.

Definition fim  := F.fim.
Definition fim1 := F.fim1.
Definition fim2 := F.fim2.
Definition at_least := F.at_least.
Definition fim3 := F.fim3.
Definition fim4 := F.fim4.
Definition fim5 := FO.fim5.
                     
Definition result0 := fim least vss os.

Definition result1 := fim1 least vss os.

Definition result2 := fim2 least vss [] os.

Definition result3 := at_least least vss (fim3 least vss [] os).

Definition result4 := fim4 least vss os.

Definition result5 := fim5 least vss os.

Time Eval vm_compute in result4.

Time Eval vm_compute in result5.

Close Scope Z_scope.