Require Import Coq.Lists.List. Import ListNotations.
Set Implicit Arguments.
Generalizable All Variables.

(*-----------------------------------------------------------------*)

(** * Algebraic Properties Defined as Coq Typeclasses *)

(** ** Unit *)
Class LeftNeutral A B (op: B -> A -> A) (e : B) :=
  {
    left_neutral  : forall a, op e a = a
  }. 

Class RightNeutral A B (op: A -> B -> A) (e : B) :=
  {
    right_neutral : forall a, op a e =  a
  }.

Class Neutral A (op: A -> A -> A) (e : A) :=
  {
    neutral_left_neutral  :>  LeftNeutral  op e;
    neutral_right_neutral :>  RightNeutral op e
  }. 

(** ** Associativity  *)
Class  Associative A (op:A->A->A) :=
  {
    associative : forall (x y z: A), op (op x y) z = op x (op y z)
  }.

(** ** Commutativity  *)
Class Commutative A (op:A->A->A) :=
  {
    commutative : forall (x y: A), op  x y = op y x
  }.

(** ** Monoid *)
Class Monoid A (op : A->A->A) (e : A) := 
  {
    monoid_assoc   :> Associative op; 
    monoid_neutral :> Neutral op e
  }.

(** ** Distributivity *)
Class Distributivity  A (otimes:A->A->A) (oplus:A->A->A) :=
  {
    left_distributivity :
      forall x y z,
        otimes x (oplus y z) =  oplus (otimes x y) (otimes x z);

    right_distributivity :
      forall x y z, 
        otimes (oplus y z) x = oplus (otimes y x) (otimes z x)
  }.

(** ** Instances *)

Instance app_associative (A:Type) : Associative (@List.app A).
Proof.
  constructor. intros.
  now rewrite app_assoc.
Qed.

Instance app_neutral (A:Type) : Neutral (@List.app A) [].
  repeat constructor; apply app_nil_r.
Qed.

Instance plus_neutral (A:Type) : Neutral plus 0.
  repeat constructor; auto with arith.
Qed.

Instance plus_associative : Associative plus.
  constructor; auto with arith.
Qed.

